#include <stdlib.h>
#include <string.h>
#include "W3RepGameData.h"

TW3RepHeader::TW3RepHeader()
{
    m_loaded = false;
    m_version = 0;
    memset(&Header, 0, sizeof(Header));
    memset(&SubHeader, 0, sizeof(SubHeader));
}

bool TW3RepHeader::Load(FILE * fp)
{
    if(!fp)
        return false;

    if(m_loaded)
        Unload();

    fread(&Header, 48, 1, fp);
    switch(Header.header_version)
    {
        case 0: fread(&SubHeader.v0, 16, 1, fp); m_version = (((WRL_DWord)SubHeader.v0.version1) << 16) + (WRL_DWord)SubHeader.v0.version2; break;
        case 1: fread(&SubHeader.v1, 20, 1, fp); m_version = (((WRL_DWord)SubHeader.v1.version1) << 16) + (WRL_DWord)SubHeader.v1.version2; break;
        default: return false;
    }

    m_loaded = true;
    return true;
}

void TW3RepHeader::Unload()
{
    if(!m_loaded)
        return;
    m_loaded = false;
    memset(&Header, 0, sizeof(Header));
    memset(&SubHeader, 0, sizeof(SubHeader));
    m_version = 0;
}

bool TW3RepHeader::Loaded() const
{
    return m_loaded;
}

const char * TW3RepHeader::GetVersionWRL_String(char * vstring, bool longversion) const
{
    if(longversion)
        sprintf(vstring, "1.%u build %u", m_version, GetBuild());
    else
        sprintf(vstring, "1.0.%u.%u", m_version>>16, m_version & 0xffff, GetBuild());
    return vstring;
}

WRL_DWord TW3RepHeader::GetVersion() const
{
    return m_version;
}

WRL_Word TW3RepHeader::GetBuild() const
{
    switch(Header.header_version)
    {
    case 0: return SubHeader.v0.buildnum;
    case 1: return SubHeader.v1.buildnum;
    }
    return 0;
}

WRL_DWord TW3RepHeader::GetLength() const
{
    switch(Header.header_version)
    {
    case 0: return SubHeader.v0.length;
    case 1: return SubHeader.v1.length;
    }
    return 0;
}

TW3RepGameData::TW3RepGameData()
{
    m_loaded = false;
    m_gamename = NULL;
    m_gameflag = 0;
    m_mapname = NULL;
    m_creator = NULL;
    m_playercount = 0;
    m_gametype = gametype_unknown;
    m_private = false;
}

TW3RepGameData::~TW3RepGameData()
{
    Unload();
}

bool TW3RepGameData::Load(WRL_String &data)
{
    if(m_loaded)
        Unload();
/*
    Read game name
*/
    m_gamename = strdup((const char *)data);
    data += (strlen(m_gamename) + 2);

/*
    Read game flag
*/
    if((*data) & 0x4)
        m_gameflag &= gameflag_hideterrain;
    if((*data) & 0x10)
        m_gameflag &= gameflag_fullshare;
    data ++;
    if((*data) & 0x2)
        m_gameflag &= gameflag_speedfast;
    data ++;
    if((*data) & 0x2)
        m_gameflag &= gameflag_mapexplored;
    if((*data) & 0x4)
        m_gameflag &= gameflag_alwaysvisible;
    if((*data) & 0x8)
        m_gameflag &= gameflag_defaultvisible;
    if((*data) & 0x10)
        m_gameflag &= gameflag_fullobserver;
    if((*data) & 0x20)
        m_gameflag &= gameflag_observerondefeat;
    if((*data) & 0x40)
        m_gameflag &= gameflag_teamtogether;
    data ++;
    if((*data) & 0x2)
        m_gameflag &= gameflag_fixedteams;
    data ++;
    if((*data) & 0x2)
        m_gameflag &= gameflag_randomhero;
    if((*data) & 0x4)
        m_gameflag &= gameflag_randomraces;
    if((*data) & 0x40)
        m_gameflag &= gameflag_referee;
    data += 11;

/*
    Read game map and creator name
*/
    DecryptInfo(data, m_mapname, m_creator);

/*
    Player Count
*/
    m_playercount = *((unsigned int *)data);
    data += 4;

/*
    Game Type
*/
    switch(*data)
    {
    case 0x1: break;
    case 0x9: m_gametype = gametype_custom; break;
    case 0x1D: m_gametype = gametype_single; break;
    case 0x20: m_gametype = gametype_team; break;
    default: m_gametype = gametype_unknown;
    }
    data ++;

/*
    Game Private Flag
*/
    if((*data) & 0x8)
        m_private = true;
    else
        m_private = false;
    data +=7;

    m_loaded = true;
    return true;
}

void TW3RepGameData::Unload()
{
    if(!m_loaded)
        return;
    m_loaded = false;
    if(m_gamename)
    {
        free((void *)m_gamename);
        m_gamename = 0;
    }
    m_gameflag = 0;
    if(m_mapname)
    {
        free((void *)m_mapname);
        m_mapname = 0;
    }
    if(m_creator)
    {
        free((void *)m_creator);
        m_creator = 0;
    }
    m_playercount = 0;
    m_gametype = gametype_unknown;
    m_private = false;
}

bool TW3RepGameData::Loaded() const
{
    return m_loaded;
}

const char * TW3RepGameData::GetGameName() const
{
    return m_gamename;
}

WRL_Word TW3RepGameData::GetGameFlag() const
{
    return m_gameflag;
}

const char * TW3RepGameData::GetMapName() const
{
    return m_mapname;
}

const char * TW3RepGameData::GetCreator() const
{
    return m_creator;
}

WRL_Byte TW3RepGameData::GetPlayerCount() const
{
    return m_playercount;
}

TGameType TW3RepGameData::GetGameType() const
{
    return m_gametype;
}

bool TW3RepGameData::GetPrivate() const
{
    return m_private;
}

bool TW3RepGameData::SetGameName(const char * gamename)
{
    if(!gamename)
        return false;
    m_gamename = strdup(gamename);
    return true;
}

bool TW3RepGameData::SetGameFlag(WRL_Word gameflag)
{
    if(m_gameflag >= 0x4000)
        return false;
    m_gameflag = gameflag;
    return true;
}

bool TW3RepGameData::SetMapName(const char * mapname)
{
    if(!mapname)
        return false;
    m_mapname = strdup(mapname);
    return true;
}

bool TW3RepGameData::SetCreator(const char * creator)
{
    if(!creator)
        return false;
    m_creator = strdup(creator);
    return true;
}

bool TW3RepGameData::SetPlayerCount(WRL_Byte playercount)
{
    if(playercount > 12)
        return false;
    m_playercount = playercount;
    return true;
}

bool TW3RepGameData::SetGameType(TGameType gametype)
{
    if(gametype > 4)
        return false;
    m_gametype = gametype;
    return true;
}

void TW3RepGameData::SetPrivate(bool priv)
{
    m_private = priv;
}

bool TW3RepGameData::DecryptInfo(WRL_String &enc, const char * &mapname, const char * &creator) const
{
    char *dec;
    unsigned pos;
    WRL_Byte bitmask = 0xff;
    int count = 0;

    if (!(mapname = strdup((const char *)enc)))
        return false;

    dec = (char *)mapname;
    pos = 0;
    while(*enc)
    {
        if ((pos + 7) % 8)
        {
            *dec = *enc;
            if (!(bitmask & 0x1)) (*dec)--;
            if((*dec) == 0 && count == 0)
            {
                if(!(creator = strdup((const char *)(enc + 1))))
                    return false;
                dec = (char *)creator;
                count ++;
            }
            else
                dec++;
            bitmask >>= 1;
        } else
            bitmask = *enc >> 1;
        enc++;
        pos++;
    }
    *dec = '\0';
    enc ++;
    return true;
}